//
//  NSImage-Extras.m
//  SimplePicture
//
//  Created by Scott Stevenson on 9/28/07.
//  Released under a BSD-style license. See License.txt

#import "NSImage-Extras.h"


@implementation NSImage (Extras)

- (NSImage*)imageByScalingProportionallyToSize:(NSSize)targetSize
{
    return [self imageByScalingProportionallyToSize:targetSize flipped:NO];
}

- (NSImage*)imageByScalingProportionallyToSize:(NSSize)targetSize flipped:(BOOL)isFlipped
{
    BOOL shouldAddShadow = YES;
    BOOL shouldAddFrame = YES;
    NSSize shadowTargetSize = targetSize;

    NSImage* sourceImage = self;
    NSImage* newImage = nil;

    if ( shouldAddShadow )
    {
        shadowTargetSize.width  -= 8.0;
        shadowTargetSize.height -= 8.0;
    }
    
    if ([sourceImage isValid])
    {
        NSSize imageSize = [sourceImage size];
        float width  = imageSize.width;
        float height = imageSize.height;

        float targetWidth  = targetSize.width;
        float targetHeight = targetSize.height;

        // scaleFactor will be the fraction that we'll
        // use to adjust the size. For example, if we shrink
        // an image by half, scaleFactor will be 0.5. the
        // scaledWidth and scaledHeight will be the original,
        // multiplied by the scaleFactor.
        //
        // IMPORTANT: the "targetHeight" is the size of the space
        // we're drawing into. The "scaledHeight" is the height that
        // the image actually is drawn at, once we take into
        // account the ideal of maintaining proportions

        float scaleFactor  = 0.0;                
        float scaledWidth  = targetWidth;
        float scaledHeight = targetHeight;

        NSPoint thumbnailPoint = NSMakePoint(0,0);

        // since not all images are square, we want to scale
        // proportionately. To do this, we find the longest
        // edge and use that as a guide.

        if ( NSEqualSizes( imageSize, targetSize ) == NO )
        {            
            // use the longeset edge as a guide. if the
            // image is wider than tall, we'll figure out
            // the scale factor by dividing it by the
            // intended width. Otherwise, we'll use the
            // height.
            
            float widthFactor;
            float heightFactor;
            
            if ( shouldAddShadow ) {
                widthFactor  = shadowTargetSize.width / width;
                heightFactor = shadowTargetSize.height / height;                                
            } else {
                widthFactor  = targetWidth / width;
                heightFactor = targetHeight / height;                                
            }
            
            if ( widthFactor < heightFactor )
                scaleFactor = widthFactor;
            else
                scaleFactor = heightFactor;

            // ex: 500 * 0.5 = 250 (newWidth)
            
            scaledWidth  = width  * scaleFactor;
            scaledHeight = height * scaleFactor;

            // center the thumbnail in the frame. if
            // wider than tall, we need to adjust the
            // vertical drawing point (y axis)
            
            if ( widthFactor < heightFactor )
                thumbnailPoint.y = (targetHeight - scaledHeight) * 0.5;

            else if ( widthFactor > heightFactor )
                thumbnailPoint.x = (targetWidth - scaledWidth) * 0.5;
  
        }

        // create a new image to draw into        
        newImage = [[NSImage alloc] initWithSize:targetSize];
        [newImage setFlipped:YES];

        // once focus is locked, all drawing goes into this NSImage instance
        // directly, not to the screen. It also receives its own graphics
        // context.
        //
        // Also, keep in mind that we're doing this in a background thread.
        // You only want to draw to the screen in the main thread, but
        // drawing to an offscreen image is (apparently) okay.

        [newImage lockFocus];
        
            [[NSGraphicsContext currentContext] setImageInterpolation:NSImageInterpolationHigh];
            NSRect thumbnailRect;
            thumbnailRect.origin = thumbnailPoint;
            thumbnailRect.size.width = scaledWidth;
            thumbnailRect.size.height = scaledHeight;
            
            if ( shouldAddShadow )
            {
                if ( height > width )
                    thumbnailRect.origin.y += 8.0; // flipped

                NSShadow* shadow = [[NSShadow alloc] init];
                [shadow setShadowColor:[NSColor colorWithCalibratedWhite:0.0 alpha:0.5]];
                [shadow setShadowOffset:NSMakeSize(4.0,4.0)];
                [shadow setShadowBlurRadius:4.0];
                                
                [NSGraphicsContext saveGraphicsState];

                    [shadow set];
                    [[NSColor whiteColor] set];
                    [NSBezierPath fillRect:thumbnailRect];                
                    
                [NSGraphicsContext restoreGraphicsState];
                
                [shadow release];                
                
            }
                
            
            if ( shouldAddShadow )
            {
                [sourceImage drawInRect: thumbnailRect
                               fromRect: NSZeroRect
                              operation: NSCompositeSourceOver
                               fraction: 1.0];                
            } else {
                
            }

                        
            if ( shouldAddFrame )
            {   
                NSRect frameRect = thumbnailRect;
                

                [[NSColor whiteColor] set];
                frameRect = NSInsetRect(frameRect,3,3);
                NSBezierPath* path2 = [NSBezierPath bezierPathWithRect:frameRect];                
                [path2 setLineWidth:6.0];
                [path2 stroke];

                [[NSColor colorWithCalibratedWhite:0.60 alpha:1.0] set];
                                NSBezierPath* path1 = [NSBezierPath bezierPathWithRect:thumbnailRect];

                [NSGraphicsContext saveGraphicsState];
                    
                    NSGraphicsContext* g = [NSGraphicsContext currentContext];
                    [g setShouldAntialias:NO];
                    [path1 setLineWidth:1.0];
                    [path1 stroke];
                
                [NSGraphicsContext restoreGraphicsState];
                

            }
            
            if ( NO )
            {
                [[NSColor redColor] set];
                [NSBezierPath strokeRect:NSMakeRect(0,0,targetSize.width,targetSize.height)];
                
            }



        [newImage unlockFocus];
        
    }

    return [newImage autorelease];
}

@end
